from typing import List, Optional
from pathlib import Path
from pydantic import BaseModel, Field


class BaseSchema(BaseModel):
    class Config:
        orm_mode = True

class DataModule(BaseSchema):
    path: str = Field(..., description="where to place the data module relative to opa data root")
    data: str = Field(..., description="data module file contents (json)")

class RegoModule(BaseSchema):
    path: str = Field(..., description="path of policy module on disk, will be used to generate policy id")
    package_name: str = Field(..., description="opa module package name")
    rego: str = Field(..., description="rego module file contents (text)")

class DeletedFiles(BaseSchema):
    data_modules: List[Path] = []
    policy_modules: List[Path] = []

class PolicyBundle(BaseSchema):
    manifest: List[str]
    hash: str = Field(..., description="commit hash (debug version)")
    old_hash: Optional[str] = Field(None, description="old commit hash (in diff bundles)")
    data_modules: List[DataModule]
    policy_modules: List[RegoModule]
    deleted_files: Optional[DeletedFiles]


from typing import IO, Generator, Callable, Optional, List, Set
from pathlib import Path

from git import Repo
from git.diff import DiffIndex, Diff
from git.objects.commit import Commit

from opal_common.git.commit_viewer import VersionedFile
from opal_common.paths import PathUtils

DiffFilter = Callable[[Diff], bool]
PathFilter = Callable[[Path], bool]

def apply_filter(
    generator: Generator[Diff, None, None],
    filter: Optional[DiffFilter] = None
) -> Generator[Diff, None, None]:
    """
    applies an optional filter on top of a Diff generator.

    returns only the diffs yielded by the source generator that pass the filter.
    if no filter is provided, returns the same results as the source generator.
    """
    if filter is None:
        yield from generator
    else:
        for diff in generator:
            if filter(diff):
                yield diff

def diffed_file_has_extension(
    diff: Diff,
    extensions: Optional[List[str]] = None
) -> bool:
    """
    filter on git diffs, filters only diffs on files that has a certain
    extension/type. if the file is renamed/added/removed, its enough
    that only one of its versions has the required extension.
    """
    if extensions is None:
        return True # no filter

    for path in [diff.a_path, diff.b_path]:
        if path is not None and Path(path).suffix in extensions:
            return True
    return False

def diffed_file_is_under_directories(diff: Diff, directories: Set[Path]) -> bool:
    """
    filter on git diffs, filters only diffs on files that are located in
    certain directories. if a file is renamed/added/removed, its enough that
    only one of its versions was located in one of the required directories.
    """
    for path in [diff.a_path, diff.b_path]:
        if path is not None and PathUtils.is_child_of_directories(Path(path), directories):
            return True
    return False

class DiffViewer:
    """
    This class allows us to view the changes made between two commits.
    these two commits are not necessarily consecutive.
    """
    def __init__(self, old: Commit, new: Commit):
        """[summary]

        Args:
            old (Commit): the older/earlier commit that defines the diff
            new (Commit): the newer/later commit that defines the diff
        """
        if old.repo != new.repo:
            raise ValueError("you can only diff two commits from the same repo!")
        self._repo: Repo = old.repo
        self._old = old
        self._new = new
        self._diffs: DiffIndex = old.diff(new)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        pass

    def changes(self, filter: Optional[DiffFilter] = None) -> Generator[Diff, None, None]:
        """
        a generator yielding all the diffs between the old commit
        and the new commit, after applying the filter.

        Each diff (instance of `Diff`) is a change in one file, i.e: if the
        diff between the commits returned by git diff is as following:

        Changes to be committed:
        (use "git restore --staged <file>..." to unstage)
            modified:   server/main.py
            deleted:    server/policy/publisher.py
            new file:   server/publisher.py

        then 3 Diffs will be returned (one modified, one deleted, one added).

        Args:
            filter (Optional[DiffFilter]): an optional predicate to filter only specific diffs.

        Yields:
            the next diff found (only for diffs passing the filter).
        """
        for diff in self._diffs:
            if filter is None:
                yield diff
            elif filter(diff):
                yield diff

    def added(self, filter: Optional[DiffFilter] = None) -> Generator[Diff, None, None]:
        """
        a generator yielding all the diffs between the old commit
        and the new commit, that are of type "new file" (i.e: added),
        after applying the filter.

        @see `changes()`
        """
        diff_generator = self._diffs.iter_change_type("A")
        yield from apply_filter(diff_generator, filter)

    def deleted(self, filter: Optional[DiffFilter] = None) -> Generator[Diff, None, None]:
        """
        a generator yielding all the diffs between the old commit
        and the new commit, that are of type "deleted", after applying the filter.

        @see `changes()`
        """
        diff_generator = self._diffs.iter_change_type("D")
        yield from apply_filter(diff_generator, filter)

    def renamed(self, filter: Optional[DiffFilter] = None) -> Generator[Diff, None, None]:
        """
        a generator yielding all the diffs between the old commit
        and the new commit, that are of type "renamed", after applying the filter.

        @see `changes()`
        """
        diff_generator = self._diffs.iter_change_type("R")
        yield from apply_filter(diff_generator, filter)

    def modified(self, filter: Optional[DiffFilter] = None) -> Generator[Diff, None, None]:
        """
        a generator yielding all the diffs between the old commit
        and the new commit, that are of type "modified", after applying the filter.

        @see `changes()`
        """
        diff_generator = self._diffs.iter_change_type("M")
        yield from apply_filter(diff_generator, filter)


    def added_files(self, filter: Optional[DiffFilter] = None) -> Generator[VersionedFile, None, None]:
        """
        a generator yielding the new version (blob) of files that were added
        (or renamed, meaning a new file was added under the new name) in the
        diff, between the old and new commits. In both cases, a new file that
        has not existed before was added to the repo.
        """
        for diff in self.added(filter):
            yield VersionedFile(diff.b_blob, self._new)

        for diff in self.renamed(filter):
            yield VersionedFile(diff.b_blob, self._new)

    def deleted_files(self, filter: Optional[DiffFilter] = None) -> Generator[VersionedFile, None, None]:
        """
        a generator yielding the old version (blob) of the files that were removed
        (or renamed, meaning the file under the old name was removed) in the diff
        between the old and new commits. In both cases, a file is removed from the repo.
        """
        for diff in self.deleted(filter):
            yield VersionedFile(diff.a_blob, self._old)

        for diff in self.renamed(filter):
            yield VersionedFile(diff.a_blob, self._old)

    def modified_files(self, filter: Optional[DiffFilter] = None) -> Generator[VersionedFile, None, None]:
        """
        a generator yielding the new version (blob) of files that were
        changed (modified) in the diff between the old and new commit.
        """
        for diff in self.modified(filter):
            yield VersionedFile(diff.b_blob, self._new)

    def added_or_modified_files(self, filter: Optional[DiffFilter] = None) -> Generator[VersionedFile, None, None]:
        """
        a shortcut generator yield both `added_files()` and `modified_files()`.
        """
        yield from self.added_files(filter)
        yield from self.modified_files(filter)

    def affected_paths(self, filter: Optional[PathFilter] = None) -> Set[Path]:
        """
        returns the set of paths of all files that were affected in the diff between
        the old and new commits. only file paths are returned (and not directories).
        """
        paths = set()
        for diff in self._diffs:
            diff: Diff
            for str_path in [diff.a_path, diff.b_path]:
                if str_path is not None:
                    path = Path(str_path)
                    if filter is None:
                        paths.add(path)
                    elif filter(path):
                        paths.add(path)
        return paths



"""
From https://github.com/orweis/emport
"""

import collections
import os
import glob
import inspect
import sys
import collections

__author__ = 'orw'


class ObjectUtils(object):

    @staticmethod
    def is_derived_of(obj, possible_parent_class):
        if hasattr(obj, "__bases__"):
            return possible_parent_class in inspect.getmro(obj)
        else:
            return False

    @staticmethod
    def get_properties(obj):
        def filter(x):
            return not isinstance(x, collections.Callable)
        return {k: v for k, v in inspect.getmembers(obj, filter) if not k.startswith("__")}

    @staticmethod
    def get_members_who_are_instance_of(obj, class_type):
        def filter(x):
            return isinstance(x, class_type)
        return inspect.getmembers(obj, filter)

    @classmethod
    def get_class_members_who_derive_of(cls, obj, parent_class):
        def filter(x):
            return inspect.isclass(x) and cls.is_derived_of(x, parent_class)\
                   and list(inspect.getmro(x)).index(parent_class) != 0
        return inspect.getmembers(obj, filter)


class PyFrame(object):

    def __init__(self):
        self._frame = inspect.currentframe()

    def __enter__(self):
        return self._frame.f_back

    def __exit__(self, exc_type, exc_value, traceback):
        del self._frame


class Emport(object):

    def __init__(self, module, members):
        self.__original__ = module
        self._members = []
        for member in members:
            self._members.append(member[1])
            setattr(self, member[0], member[1])

    def get_original_module(self):
        return self.__original__

    def get_members_list(self):
        return self._members

    def get_flat_list(self):
        """
        :return: all the members of this Emport (And submodules) as one list
        """
        res = []
        for member in self._members:

            # if a member is an Emport itself flatten it as well
            if isinstance(member, Emport):
                res += member.get_flat_list()
            else:
                res.append(member)
        return res

    def __repr__(self):
        return "EMPORT - %s" % self.__original__


def get_caller_module(depth=0):
    """
    :param depth: stack depth of the caller. 0 == yourself, 1 == your parent
    :return: the module object of the caller function (in set stack depth)
    """
    with PyFrame() as frame:
        for i in range(0, depth):
            frame = frame.f_back
        return sys.modules[frame.f_globals["__name__"]]


def co_to_dict(co):
    return {'co_argcount': co.co_argcount, 'co_nlocals': co.co_nlocals, 'co_stacksize': co.co_stacksize,
            'co_flags': co.co_flags, 'co_consts': co.co_consts, 'co_names': co.co_names,
            'co_varnames': co.co_varnames, 'co_filename': co.co_filename,
            'co_name': co.co_name, 'co_firstlineno': co.co_firstlineno, 'co_lnotab': co.co_lnotab}


def get_caller(depth=0):
    """
    :param depth: stack depth of the caller. 0 == yourself, 1 == your parent
    :return: the frame object of the caller function (in set stack depth)
    """
    with PyFrame() as frame:
        for i in range(0, depth):
            frame = frame.f_back
        return co_to_dict(frame.f_code)



def emport_by_class(from_path, cls, import_items=None):
    """
    Wrap __import__ to import modules and filter only classes deriving from the given cls
    :param from_path: dot separated package path
    :param cls: class to filter import contents by
    :param import_items: the items to import form the package path (can also be ['*'])
    :return: an Emport object with contents filtered according to given cls
    """
    import_items = import_items or ["*"]
    module_obj = __import__(from_path, globals(), locals(), import_items, 0)
    clean_items = ObjectUtils.get_class_members_who_derive_of(module_obj, cls)
    for (sub_name, sub_module) in ObjectUtils.get_members_who_are_instance_of(module_obj, module_obj.__class__):
        results = ObjectUtils.get_class_members_who_derive_of(sub_module, cls)
        # Keep only modules with sub values
        if len(results) > 0:
            clean_sub_module = Emport(sub_module, results)
            clean_items.append((sub_name, clean_sub_module))
    clean_module = Emport(module_obj, clean_items)
    return clean_module

def emport_objects_by_class(from_path, cls, import_items=None):
    """
    Wrap __import__ to import modules and filter only classes deriving from the given cls
    Return a flat list of objects without the modules themselves
    :param from_path: dot separated package path
    :param cls: class to filter import contents by
    :param import_items: the items to import form the package path (can also be ['*'])
    :return: an Emport object with contents filtered according to given cls
    """
    results = []
    import_items = import_items or ["*"]
    module_obj = __import__(from_path, globals(), locals(), import_items, 0)
    # direct objects
    clean_items = ObjectUtils.get_class_members_who_derive_of(module_obj, cls)
    results.extend(clean_items)
    # nested
    for (sub_name, sub_module) in ObjectUtils.get_members_who_are_instance_of(module_obj, module_obj.__class__):
        objects = ObjectUtils.get_class_members_who_derive_of(sub_module, cls)
        results.extend(objects)
    return results

def dynamic_all(init_file_path):
    """
    return a list of all the py files in a dir
    usage (in __init__.py file) :
        from emport import dynamic_all
        __all__ = dynamic_all(__file__)
    """
    modules = glob.glob(os.path.join(os.path.dirname(init_file_path), "*.py*"))
    target_modules = set([])
    for module in modules:
        name = os.path.splitext(os.path.basename(module))[0]
        if os.path.isfile(module) and not name.startswith("_"):
            target_modules.add(name)
    return list(target_modules)


import os

from functools import partial
from typing import Optional
from pathlib import Path
from tenacity import retry, wait_fixed, stop_after_attempt, RetryError
from git import Repo, GitError, GitCommandError

from opal_common.logger import logger
from opal_common.git.exceptions import GitFailed
from opal_common.config import GIT_SSH_KEY_FILE


SSH_PREFIX = "ssh://"
GIT_SSH_USER_PREFIX = "git@"

def is_ssh_repo_url(repo_url: str):
    """
    return True if the repo url uses SSH authentication.
    (see: https://docs.github.com/en/github/authenticating-to-github/connecting-to-github-with-ssh)
    """
    return repo_url.startswith(SSH_PREFIX) or repo_url.startswith(GIT_SSH_USER_PREFIX)


class CloneResult:
    """
    wraps a git.Repo instance but knows if the repo was initialized with a url
    and cloned from a remote repo, or was initialed from a local `.git` repo.
    """
    def __init__(self, repo: Repo, cloned_from_remote: bool):
        self._repo = repo
        self._cloned_from_remote = cloned_from_remote

    @property
    def repo(self) -> Repo:
        """
        the wrapped repo instance
        """
        return self._repo

    @property
    def cloned_from_remote(self) -> bool:
        """
        whether the repo was cloned from remote, or we found a local matching .git repo
        """
        return self._cloned_from_remote


class RemoteClone(CloneResult):
    def __init__(self, repo: Repo):
        super().__init__(repo=repo, cloned_from_remote=True)


class LocalClone(CloneResult):
    def __init__(self, repo: Repo):
        super().__init__(repo=repo, cloned_from_remote=False)


class RepoCloner:
    """
    simple wrapper for git.Repo() to simplify other classes that need to deal
    with the case where a repo must be cloned from url *only if* the repo does
    not already exists locally, and otherwise initialize the repo instance from
    the repo already existing on the filesystem.
    """

    DEFAULT_RETRY_CONFIG = {
        'wait': wait_fixed(5),
        'stop': stop_after_attempt(2),
        'reraise': True,
    }

    def __init__(
        self,
        repo_url: str,
        clone_path: str,
        retry_config = None,
        ssh_key: Optional[str] = None,
        ssh_key_file_path: Optional[str] = GIT_SSH_KEY_FILE,
    ):
        """inits the repo cloner.

        Args:
            repo_url (str): the url to the remote repo we want to clone
            clone_path (str): the target local path in our file system we want the
                repo to be cloned to
            retry_config (dict): Tenacity.retry config (@see https://tenacity.readthedocs.io/en/latest/api.html#retry-main-api)
            ssh_key (str, optional): private ssh key used to gain access to the cloned repo
            ssh_key_file_path (str, optional): local path to save the private ssh key contents
        """
        if repo_url is None:
            raise ValueError("must provide repo url!")

        self.url = repo_url
        self.path = os.path.expanduser(clone_path)
        self._ssh_key = ssh_key
        self._ssh_key_file_path = ssh_key_file_path
        self._retry_config = retry_config if retry_config is not None else self.DEFAULT_RETRY_CONFIG

    def clone(self) -> CloneResult:
        """
        initializes a git.Repo and returns the clone result.
        it either:
            - does not found a cloned repo locally and clones from remote url
            - finds a cloned repo locally and does not clone from remote.
        """
        logger.info("Cloning repo from '{url}' to '{to_path}'", url=self.url, to_path=self.path)
        git_path = Path(self.path) / Path(".git")
        if git_path.exists():
            return self._attempt_init_from_local_repo()
        else:
            return self._attempt_clone_from_url()

    def _attempt_init_from_local_repo(self) -> CloneResult:
        """
        inits the repo from local .git or throws GitFailed
        """
        logger.info("Repo already exists in '{repo_path}'", repo_path=self.path)
        try:
            repo = Repo(self.path)
        except Exception as e:
            logger.exception("cannot init local repo: {error}", error=e)
            raise GitFailed(e)

        return LocalClone(repo)

    def _attempt_clone_from_url(self) -> CloneResult:
        """
        clones the repo from url or throws GitFailed
        """
        env = self._provide_git_ssh_environment()
        _clone_func = partial(Repo.clone_from, url=self.url, to_path=self.path, env=env)
        _clone_with_retries = retry(**self._retry_config)(_clone_func)
        try:
            repo = _clone_with_retries()
        except (GitError, GitCommandError) as e:
            logger.exception("cannot clone policy repo: {error}", error=e)
            raise GitFailed(e)
        except RetryError as e:
            logger.exception("cannot clone policy repo: {error}", error=e)
            raise GitFailed(e)
        else:
            logger.info("Clone succeeded", repo_path=self.path)
            return RemoteClone(repo)

    def _provide_git_ssh_environment(self):
        """
        provides git SSH configuration via GIT_SSH_COMMAND.

        the git ssh config will be provided only if the following conditions are met:
        - the repo url is a git ssh url
        - an ssh private key is provided in Repo Cloner __init__
        """
        if not is_ssh_repo_url(self.url) or self._ssh_key is None:
            return None # no ssh config
        git_ssh_identity_file = self._save_ssh_key_to_pem_file(self._ssh_key)
        return {
            "GIT_SSH_COMMAND": f"ssh -o StrictHostKeyChecking=no -i {git_ssh_identity_file}"
        }

    def _save_ssh_key_to_pem_file(self, key: str) -> Path:
        key = key.replace("_", "\n")
        if not key.endswith('\n'):
            key = key + '\n' # pem file must end with newline
        key_path = os.path.expanduser(self._ssh_key_file_path)
        parent_directory = os.path.dirname(key_path)
        if not os.path.exists(parent_directory):
            os.makedirs(parent_directory, exist_ok=True)
        with open(key_path, 'w') as f:
            f.write(key)
        os.chmod(key_path, 0o600)
        return Path(key_path)



from pathlib import Path
from .confi import Confi

confi = Confi(prefix="OPAL_")

ALLOWED_ORIGINS = confi.list("ALLOWED_ORIGINS", ["*"])
PROCESS_NAME = ""
# Logging
# - Log formatting
LOG_TRACEBACK = confi.bool("LOG_TRACEBACK", True)
LOG_DIAGNOSE = confi.bool("LOG_DIAGNOSE", True)
LOG_COLORIZE = confi.bool("LOG_COLORIZE", True)
LOG_SHOW_CODE_LINE = confi.bool("LOG_SHOW_CODE_LINE", True)
#  - log level
LOG_LEVEL = confi.str("LOG_LEVEL", "INFO")
#  - Which modules should be logged
LOG_MODULE_EXCLUDE_LIST = confi.list("LOG_MODULE_EXCLUDE_LIST", [
    "fastapi_websocket_rpc",
    "fastapi_websocket_pubsub",
    "uvicorn",
    # NOTE: the env var LOG_MODULE_EXCLUDE_OPA affects this list
])
LOG_MODULE_INCLUDE_LIST = confi.list("LOG_MODULE_INCLUDE_LIST", ["uvicorn.protocols.http"])
LOG_PATCH_UVICORN_LOGS = confi.bool("LOG_PATCH_UVICORN_LOGS", True,
                                    description="Should we takeover UVICORN's logs so they appear in the main logger")
# - Log to file as well ( @see https://github.com/Delgan/loguru#easier-file-logging-with-rotation--retention--compression)
LOG_TO_FILE = confi.bool("LOG_TO_FILE", False, description="Should we log to a file")
LOG_FILE_PATH = confi.str("LOG_FILE_PATH", f"opal_{PROCESS_NAME}{{time}}.log", description="path to save log file")
LOG_FILE_ROTATION = confi.str("LOG_FILE_ROTATION", "250 MB")
LOG_FILE_RETENTION = confi.str("LOG_FILE_RETENTION", "10 days")
LOG_FILE_COMPRESSION = confi.str("LOG_FILE_COMPRESSION", None)
LOG_FILE_SERIALIZE = confi.str("LOG_FILE_SERIALIZE", True)
LOG_FILE_LEVEL = confi.str("LOG_FILE_LEVEL", "INFO")

# Fetching Providers
# - where to load providers from
FETCH_PROVIDER_MODULES = confi.list("FETCH_PROVIDER_MODULES", ["opal_common.fetcher.providers"])

GIT_SSH_KEY_FILE = confi.str("GIT_SSH_KEY_FILE", str(Path.home() / ".ssh/opal_repo_ssh_key"))


from loguru import logger
import logging
import sys
from .logging.intercept import InterceptHandler
from .logging.thirdparty import hijack_uvicorn_logs
from .logging.formatter import Formatter
from .logging.filter import ModuleFilter
from . import config


def configure_logs():
    """
    Takeover process logs and create a logger with Loguru according to the configuration
    """
    intercept_handler = InterceptHandler()
    formatter = Formatter()
    filter = ModuleFilter(include_list=config.LOG_MODULE_INCLUDE_LIST, exclude_list=config.LOG_MODULE_EXCLUDE_LIST)
    logging.basicConfig(handlers=[intercept_handler], level=0, force=True)
    if config.LOG_PATCH_UVICORN_LOGS:
        # Monkey patch UVICORN to use our logger
        hijack_uvicorn_logs()
    # Clean slate
    logger.remove()
    # Logger configuration
    logger.add(
        sys.stderr,
        filter=filter.filter,
        format=formatter.format,
        level=config.LOG_LEVEL,
        backtrace=config.LOG_TRACEBACK,
        diagnose=config.LOG_DIAGNOSE,
        colorize=config.LOG_COLORIZE,
    )
    # log to a file
    if config.LOG_TO_FILE:
        logger.add(
            config.LOG_FILE_PATH,
            compression=config.LOG_FILE_COMPRESSION,
            retention=config.LOG_FILE_RETENTION,
            rotation=config.LOG_FILE_ROTATION,
            serialize=config.LOG_FILE_SERIALIZE,
            level=config.LOG_FILE_LEVEL,
        )


def get_logger(name=""):
    """
    backward comptability to old get_logger
    """
    return logger


configure_logs()



from fastapi import FastAPI, status, Request
from fastapi.responses import JSONResponse
from fastapi.encoders import jsonable_encoder
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel

from opal_common.config import ALLOWED_ORIGINS
from opal_common.logger import logger

class ErrorResponse(BaseModel):
    error: str


def get_response() -> JSONResponse:
    error = ErrorResponse(error="Uncaught server exception")
    json_error = jsonable_encoder(error.dict())
    return JSONResponse(content=json_error, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)


def register_default_server_exception_handler(app: FastAPI):
    """
    Registers a default exception handler for HTTP 500 exceptions.
    Since fastapi does not include CORS headers by default in 500 exceptions,
    we need to include them manually. Otherwise the frontend cries on the wrong issue.
    """
    @app.exception_handler(status.HTTP_500_INTERNAL_SERVER_ERROR)
    async def default_server_exception_handler(request: Request, exception: Exception):
        response = get_response()
        logger.exception("Uncaught server exception: {exc}", exc=exception)

        # Since the CORSMiddleware is not executed when an unhandled server exception
        # occurs, we need to manually set the CORS headers ourselves if we want the FE
        # to receive a proper JSON 500, opposed to a CORS error.
        # Setting CORS headers on server errors is a bit of a philosophical topic of
        # discussion in many frameworks, and it is currently not handled in FastAPI.
        # See dotnet core for a recent discussion, where ultimately it was
        # decided to return CORS headers on server failures:
        # https://github.com/dotnet/aspnetcore/issues/2378
        origin = request.headers.get('origin')

        if origin:
            # Have the middleware do the heavy lifting for us to parse
            # all the config, then update our response headers
            cors = CORSMiddleware(
                app=app,
                allow_origins=ALLOWED_ORIGINS,
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"])

            # Logic directly from Starlette's CORSMiddleware:
            # https://github.com/encode/starlette/blob/master/starlette/middleware/cors.py#L152

            response.headers.update(cors.simple_headers)
            has_cookie = "cookie" in request.headers

            # If request includes any cookie headers, then we must respond
            # with the specific origin instead of '*'.
            if cors.allow_all_origins and has_cookie:
                response.headers["Access-Control-Allow-Origin"] = origin

            # If we only allow specific origins, then we have to mirror back
            # the Origin header in the response.
            elif not cors.allow_all_origins and cors.is_allowed_origin(origin=origin):
                response.headers["Access-Control-Allow-Origin"] = origin
                response.headers.add_vary_header("Origin")

        return response

def configure_cors_middleware(app: FastAPI):
    app.add_middleware(
        CORSMiddleware,
        allow_origins=ALLOWED_ORIGINS,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

def configure_middleware(app: FastAPI):
    register_default_server_exception_handler(app)
    configure_cors_middleware(app)



from .events import FetchEvent
from tenacity import retry, wait, stop
import tenacity

from .logger import get_logger
logger = get_logger("opal.providers")


class BaseFetchProvider:
    """
    Base class for data fetching providers.
     - Override self._fetch_ to implement fetching
     - call self.fetch() to retrive data (wrapped in retries and safe execution guards)
     - override __aenter__ and __aexit__ for async context 
    """

    DEFAULT_RETRY_CONFIG = {
        'wait': wait.wait_random_exponential(),
        "stop": stop.stop_after_attempt(200),
        "reraise": True
    }

    def __init__(self, event: FetchEvent, retry_config=None) -> None:
        """
        Args:
            event (FetchEvent): the event desciring what we should fetch
            retry_config (dict): Tenacity.retry config (@see https://tenacity.readthedocs.io/en/latest/api.html#retry-main-api) for retrying fetching
        """
        # convert the event as needed and save it
        self._event = self.parse_event(event)
        self._url = event.url
        self._retry_config = retry_config if retry_config is not None else self.DEFAULT_RETRY_CONFIG

    def parse_event(self, event:FetchEvent)->FetchEvent:
        """
        Parse the event (And config within it) into the right object type

        Args:
            event (FetchEvent): the event to be parsed

        Returns:
            FetchEvent: an event deriving from FetchEvent
        """
        return event
        

    async def fetch(self):
        """
        Fetch and return data.
        Calls self._fetch_ with a retry mechanism
        """
        attempter = retry(**self._retry_config)(self._fetch_)
        res = await attempter()
        return res

    async def process(self, data):
        try: 
            return await self._process_(data)
        except:
            logger.exception("Failed to process fetched data")
            raise

    async def __aenter__(self):
        return self 

    async def __aexit__(self, exc_type=None, exc_val=None, tb=None):
        pass

    async def _fetch_(self):
        """
        Internal fetch operation called by self.fetch()
        Override this method to implement a new fetch provider
        """
        pass

    async def _process_(self, data):
        return data

    def set_retry_config(self, retry_config:dict):
        """ 
        Set the configuration for retrying failed fetches
        @see self.DEFAULT_RETRY_CONFIG

        Args:
            retry_config (dict): Tenacity retry config
        """
        self._retry_config = retry_config





from typing import Coroutine

from ..fetcher_register import FetcherRegister
from ..events import FetcherConfig, FetchEvent
from .core_callbacks import OnFetchFailureCallback



class BaseFetchingEngine:
    """
    An interface base class for a task queue manager used for fetching events.
    """

    @property
    def register(self)->FetcherRegister:
        """
        access to the underlying fetcher providers register
        """
        raise NotImplementedError()   

    async def __aenter__(self):
        """
        Async Context manager to cancel tasks on exit 
        """
        raise NotImplementedError()

    async def __aexit__(self, exc_type, exc, tb):
        raise NotImplementedError()

    async def terminate_tasks(self):
        """
        Cancel and wait on the internal worker tasks
        """
        raise NotImplementedError()

    async def queue_url(self, url: str, callback: Coroutine, config: FetcherConfig = None, fetcher="HttpGetFetchProvider") -> FetchEvent:
        """
        Simplified default fetching handler for queuing a fetch task

        Args:
            url (str): the URL to fetch from
            callback (Coroutine): a callback to call with the fetched result
            config (FetcherConfig, optional): Configuration to be used by the fetcher. Defaults to None.
            fetcher (str, optional): Which fetcher class to use. Defaults to "HttpGetFetchProvider".
        Returns: 
            the queued event (which will be mutated to at least have an Id)
        """
        raise NotImplementedError()

    async def queue_fetch_event(self, event: FetchEvent, callback: Coroutine) -> FetchEvent:
        """
        Basic handler to queue a fetch event for a fetcher class.
        Waits if the queue is full.

        Args:
            event (FetchEvent): the fetch event to queue as a task
            callback (Coroutine): a callback to call with the fetched result
        Returns: 
            the queued event (which will be mutated to at least have an Id)            
        """
        raise NotImplementedError()


    def register_failure_handler(self, callback: OnFetchFailureCallback):
        """
        Register a callback to be called with exception and original event in case of failure

        Args:
            callback (OnFetchFailureCallback): callback to register
        """
        raise NotImplementedError()

    async def _on_failure(self, error: Exception, event: FetchEvent):
        """
        Call event failure subscribers

        Args:
            error (Exception): thrown exception
            event (FetchEvent): event which was being handled
        """
        raise NotImplementedError()




from logging.config import dictConfig
def hijack_uvicorn_logs():
    # Intercept future UVICORN
    from uvicorn.config import LOGGING_CONFIG
    LOGGING_CONFIG["handlers"] =  {
            "default": {
                "formatter": "default",
                "class": "opal_common.logging.intercept.InterceptHandler",
            },
            "access": {
                "formatter": "access",
                "class": "opal_common.logging.intercept.InterceptHandler",
            },
        }
    # force existing UVICORN logging
    dictConfig(LOGGING_CONFIG)






from functools import partial
from pathlib import Path
from typing import List, Optional, Set

from git import Repo
from git.objects import Commit

from opal_common.git.commit_viewer import CommitViewer, has_extension, is_under_directories
from opal_common.git.diff_viewer import DiffViewer, diffed_file_has_extension, diffed_file_is_under_directories
from opal_common.opa import get_rego_package, is_data_module, is_rego_module
from opal_common.schemas.policy import DataModule, PolicyBundle, RegoModule, DeletedFiles


class BundleMaker:
    """
    creates a policy bundle based on:
    - the current state of the policy git repo
    - filtering criteria on the policy git repo (specific directories, specific file types, etc)

    there are two types of bundles:
    - a full/complete bundle, representing the state of the repo at one commit
    - a diff bundle, representing only the *changes* made to the policy between two commits (the diff).
    """
    def __init__(self, repo: Repo, in_directories: Set[Path], extensions: Optional[List[str]] = None):
        """[summary]

        Args:
            repo (Repo): the policy repo
            in_directories (Set[Path]): the directories in the repo that we want to filter on.
                if the entire repo is relevant, pass Path(".") as the directory
                (all paths are relative to the repo root).
            extensions (Optional[List[str]]): optional filtering on file extensions.
        """
        self._repo = repo
        self._directories = in_directories
        self._has_extension = partial(has_extension, extensions=extensions)
        self._is_under_directories = partial(is_under_directories, directories=in_directories)
        self._diffed_file_has_extension = partial(diffed_file_has_extension, extensions=extensions)
        self._diffed_file_is_under_directories = partial(diffed_file_is_under_directories, directories=in_directories)

    def make_bundle(self, commit: Commit) -> PolicyBundle:
        """
        creates a *complete* bundle of all the policy and data modules found
        in the policy repo, when the repo HEAD is at the given `commit`.

        Args:
            commit (Commit): the commit the repo should be checked out on to search for policy files.

        Returns:
            bundle (PolicyBundle): the bundle of policy modules found in the repo (checked out on `commit`)
        """
        data_modules = []
        policy_modules = []
        manifest = []

        with CommitViewer(commit) as viewer:
            filter = lambda f: self._has_extension(f) and self._is_under_directories(f)
            for source_file in viewer.files(filter):
                contents = source_file.read()
                path = source_file.path

                if is_data_module(path):
                    data_modules.append(
                        DataModule(path=str(path.parent), data=contents)
                    )
                elif is_rego_module(path):
                    policy_modules.append(
                        RegoModule(
                            path=str(path),
                            package_name=get_rego_package(contents) or "",
                            rego=contents,
                        )
                    )
                else:
                    continue

                manifest.append(str(path)) # only returns the relative path

            return PolicyBundle(
                manifest=manifest,
                hash=commit.hexsha,
                data_modules=data_modules,
                policy_modules=policy_modules,
            )

    def make_diff_bundle(self, old_commit: Commit, new_commit: Commit) -> PolicyBundle:
        """
        creates a *diff* bundle of all the policy and data modules that were changed
        (either added, renamed, modified or deleted) between the `old_commit` and `new_commit`.
        essentially all the relevant files when running `git diff old_commit..new_commit`.

        Note that we still filter only directories and file types given in the constructor.

        Args:
            old_commit (Commit): represents the previous known state of the repo.
                The opal client subscribes to the policy state and gets updates from
                the server via a pubsub channel. When it receives an update that
                new state is available in the policy repo, the client requests a
                *diff bundle* from the /policy api route. The client will report
                its last known commit as the `old_commit`, and only new state
                (the diff from the client known commit to the server newest commit)
                will be returned back.
            commit (Commit): represents the newest known commit in the server (the new state).

        Returns:
            bundle (PolicyBundle): a diff bundle containing only the policy modules changed
                between `old_commit` and `new_commit`.
        """
        data_modules = []
        policy_modules = []
        deleted_data_modules = []
        deleted_policy_modules = []
        manifest = []

        with DiffViewer(old_commit, new_commit) as viewer:
            filter = lambda diff: (
                self._diffed_file_has_extension(diff) and \
                    self._diffed_file_is_under_directories(diff)
                )
            for source_file in viewer.added_or_modified_files(filter):
                contents = source_file.read()
                path = source_file.path

                if is_data_module(path):
                    data_modules.append(
                        # in OPA, the data module path is the containing directory
                        # i.e: /path/to/data.json will put the json contents under "/path/to" in the opa tree
                        DataModule(path=str(path.parent), data=contents)
                    )
                elif is_rego_module(path):
                    policy_modules.append(
                        RegoModule(
                            path=str(path),
                            package_name=get_rego_package(contents) or "",
                            rego=contents,
                        )
                    )
                else:
                    continue

                manifest.append(str(path)) # only returns the relative path

            for source_file in viewer.deleted_files(filter):
                path = source_file.path

                if is_data_module(path):
                    # in OPA, the data module path is the containing directory (see above)
                    deleted_data_modules.append(str(path.parent))
                elif is_rego_module(path):
                    deleted_policy_modules.append(path)

            if deleted_data_modules or deleted_policy_modules:
                deleted_files = DeletedFiles(
                    data_modules=deleted_data_modules,
                    policy_modules=deleted_policy_modules,
                )
            else:
                deleted_files = None

            return PolicyBundle(
                manifest=manifest,
                hash=new_commit.hexsha,
                old_hash=old_commit.hexsha,
                data_modules=data_modules,
                policy_modules=policy_modules,
                deleted_files=deleted_files
            )


import asyncio
from typing import Callable, Coroutine, List, Optional
from git.objects import Commit

from opal_common.git.branch_tracker import BranchTracker
from opal_common.git.repo_cloner import RepoCloner
from opal_common.git.exceptions import GitFailed
from opal_common.logger import logger


OnNewCommitsCallback = Callable[[Commit, Commit], Coroutine]
OnGitFailureCallback = Callable[[Exception], Coroutine]


class RepoWatcher:
    """
    Watches a git repository for changes and can trigger callbacks
    when detecting new commits on the tracked branch.

    Checking for changes is done following a git pull from a tracked
    remote. The pull can be either triggered by a method (i.e: you can
    call it from a webhook) or can be triggered periodically by a polling
    task.
    """
    def __init__(
        self,
        repo_url: str,
        clone_path: str,
        branch_name: str = "master",
        remote_name: str = "origin",
        ssh_key: Optional[str] = None,
        polling_interval: int = 0,
    ):
        self._cloner = RepoCloner(repo_url, clone_path, ssh_key=ssh_key)
        self._branch_name = branch_name
        self._remote_name = remote_name
        self._tracker = None
        self._on_failure_callbacks: List[OnNewCommitsCallback] = []
        self._on_new_commits_callbacks: List[OnGitFailureCallback] = []
        self._polling_interval = polling_interval
        self._polling_task = None

    def on_new_commits(self, callback: OnNewCommitsCallback):
        """
        Register a callback that will be called when new commits
        are detected on the monitored repo (after a pull).
        """
        self._on_new_commits_callbacks.append(callback)

    def on_git_failed(self, callback: OnGitFailureCallback):
        """
        Register a callback that will be called when one of the underlying
        git actions fails without possibility for recovery, i.e: we try to
        clone from a bad repo_url, etc.
        """
        self._on_failure_callbacks.append(callback)

    async def run(self):
        """
        clones the repo and potentially starts the polling task
        """
        try:
            result = self._cloner.clone()
        except GitFailed as e:
            await self._on_git_failed(e)
            return

        self._tracker = BranchTracker(
            repo=result.repo,
            branch_name=self._branch_name,
            remote_name=self._remote_name
        )
        # if the repo exists locally, we need to git pull when the watcher starts
        if not result.cloned_from_remote:
            self._tracker.pull()

        if (self._polling_interval > 0):
            logger.info("Launching polling task, interval: {interval} seconds", interval=self._polling_interval)
            self._start_polling_task()
        else:
            logger.info("Polling task is off")

    async def stop(self):
        return await self._stop_polling_task()

    async def check_for_changes(self):
        """
        calling this method will trigger a git pull from the tracked remote.
        if after the pull the watcher detects new commits, it will call the
        callbacks registered with on_new_commits().
        """
        logger.info("Pulling changes from remote: '{remote}'", remote=self._tracker.tracked_remote.name)
        has_changes, prev, latest = self._tracker.pull()
        if not has_changes:
            logger.info("No new commits: HEAD is at '{head}'", head=latest.hexsha)
        else:
            logger.info("Found new commits: old HEAD was '{prev_head}', new HEAD is '{new_head}'", prev_head=prev.hexsha, new_head=latest.hexsha)
            await self._on_new_commits(old=prev, new=latest)

    async def _do_polling(self):
        """
        optional task to periodically check the remote for changes (git pull and compare hash).
        """
        while True:
            await asyncio.sleep(self._polling_interval)
            await self.check_for_changes()

    def _start_polling_task(self):
        if self._polling_task is None and self._polling_interval > 0:
            self._polling_task = asyncio.create_task(self._do_polling())

    async def _stop_polling_task(self):
        if self._polling_task is not None:
            self._polling_task.cancel()
            try:
                await self._polling_task
            except asyncio.CancelledError:
                pass

    async def _on_new_commits(self, old: Commit, new: Commit):
        """
        triggers callbacks registered with on_new_commits().
        """
        await self._run_callbacks(self._on_new_commits_callbacks, old, new)

    async def _on_git_failed(self, exc: Exception):
        """
        will be triggered if a git failure occurred (i.e: repo does not exist, can't clone, etc).
        triggers callbacks registered with on_git_failed().
        """
        await self._run_callbacks(self._on_failure_callbacks, exc)

    async def _run_callbacks(self, handlers, *args, **kwargs):
        """
        triggers a list of callbacks
        """
        await asyncio.gather(*(callback(*args, **kwargs) for callback in handlers))



from typing import Coroutine, Any
try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol

from fastapi_websocket_pubsub import PubSubClient, Topic, TopicList
from opal_common.logger import logger


class TopicCallback(Protocol):
    def __call__(self, topic: Topic, data: Any) -> Coroutine: ...


class TopicListener:
    """
    A simple wrapper around a PubSubClient that listens on a topic
    and runs a callback when messages arrive for that topic.
    Provides start() and stop() shortcuts that helps treat this client
    as a separate "process" or task that runs in the background.
    """
    def __init__(
        self,
        client: PubSubClient,
        server_uri: str,
        topics: TopicList = None,
        callback: TopicCallback = None,
    ):
        """[summary]

        Args:
            client (PubSubClient): a configured not-yet-started pub sub client
            server_uri (str): the URI of the pub sub server we subscribe to
            topics (TopicList): the topic(s) we subscribe to
            callback (TopicCallback): the (async) callback to run when a message
                arrive on one of the subsribed topics
        """
        self._client = client
        self._server_uri = server_uri
        self._topics = topics
        self._callback = callback

    async def __aenter__(self):
        self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()

    def start(self):
        """
        starts the pub/sub client and subscribes to the predefined topic.
        the client will attempt to connect to the pubsub server until successful.
        """
        logger.info("started topic listener, topics={topics}", topics=self._topics)
        for topic in self._topics:
            self._client.subscribe(topic, self._callback)
        self._client.start_client(f"{self._server_uri}")

    async def stop(self):
        """
        stops the pubsub client
        """
        await self._client.disconnect()
        logger.info("stopped topic listener", topics=self._topics)

    async def wait_until_done(self):
        """
        When the listener is a used as a context manager, this method waits until
        the client is done (i.e: terminated) to prevent exiting the context.
        """
        return await self._client.wait_until_done()


from typing import List


class ModuleFilter:
    """
    filter logs by module name
    """

    def __init__(self, exclude_list:List[str]=None,include_list:List[str]=None ) -> None:
        """[summary]

        Args:
            exclude_list (List[str], optional): module name (prefixes) to reject. Defaults to [].
            include_list (List[str], optional): module name (prefixes) to include (even if higher form is excluded). Defaults to [].

        Usage:
            ModuleFilter(["uvicorn"]) # exclude all logs coming from module name starting with "uvicorn"
            ModuleFilter(["uvicorn"], ["uvicorn.access]) # exclude all logs coming from module name starting with "uvicorn" except ones starting with "uvicorn.access")

        """
        self._exclude_list = exclude_list or []
        self._include_list = include_list or []

    def filter(self, record):
        name:str = record["name"]
        for module in self._include_list:
            if name.startswith(module):
                return True
        for module in self._exclude_list:
            if name.startswith(module):
                return False
        return True
        


from typing import IO, Generator, Callable, Optional, List, Set

from git import Repo
from git.objects import Commit, Tree, Blob, IndexObject
from pathlib import Path

from opal_common.paths import PathUtils

class VersionedNode:
    """
    A *versioned* file or a directory in a git repo.
    VersionedNode is a base class for `VersionedFile` and `VersionedDirectory`.
    """
    def __init__(self, node: IndexObject, commit: Commit):
        self._node = node
        self._commit = commit
        self._repo: Repo = commit.repo

    @property
    def repo(self) -> Repo:
        """
        the repo containing the versioned node
        """
        return self._repo

    @property
    def commit(self) -> Commit:
        """
        the commit in which the node (blob, tree) is located
        """
        return self._commit

    @property
    def version(self) -> str:
        """
        the hash (hex sha) of the node's parent commit
        """
        return self._commit.hexsha

    @property
    def path(self) -> Path:
        """
        the relative path to the node (either file path or directory path),
        relative to the repo root.
        """
        return Path(self._node.path)

class VersionedFile(VersionedNode):
    """
    Each instance of this class represents *one version* of a file (blob)
    in a git repo (the version of the file for a specific git commit).
    """
    def __init__(self, blob: Blob, commit: Commit):
        super().__init__(blob, commit)
        self._blob: Blob = blob

    @property
    def blob(self) -> Blob:
        """
        the blob containing metadata for the file version
        """
        return self._blob

    @property
    def stream(self) -> IO:
        """
        an io stream to the version of the file represented by that instance.
        reading that stream will return the contents of the file for that
        specific version (commit).
        """
        return self.blob.data_stream

    def read_bytes(self) -> bytes:
        """
        returns the contents of the file as a byte array (without encoding).
        """
        return self.stream.read()

    def read(self, encoding='utf-8') -> str:
        """
        returns the contents of the file as a string, decoded according to the input `encoding`.

        (by default, git usually encodes source files as utf-8).
        """
        return self.read_bytes().decode(encoding=encoding)

class VersionedDirectory(VersionedNode):
    """
    Each instance of this class represents *one version* of a directory (git tree)
    in a git repo (the version of the directory for a specific git commit).
    """
    def __init__(self, directory: Tree, commit: Commit):
        super().__init__(directory, commit)
        self._dir: Tree = directory

    @property
    def dir(self) -> Tree:
        """
        the git tree representing the metadata for that version of the directory.
        i.e: one can get child directories (trees) and files (blobs) for the instance's version.
        """
        return self._dir


NodeFilter = Callable[[VersionedNode], bool]
FileFilter = Callable[[VersionedFile], bool]
DirectoryFilter = Callable[[VersionedDirectory], bool]


def has_extension(f: VersionedFile, extensions: Optional[List[str]] = None) -> bool:
    """
    a filter on versioned files, filters only files with specific types (file extensions).
    """
    if extensions is None:
        return True # no filter
    else:
        return f.path.suffix in extensions


def is_under_directories(f: VersionedFile, directories: Set[Path]) -> bool:
    """
    a filter on versioned files, filters only files under certain directories in the repo.
    """
    return PathUtils.is_child_of_directories(f.path, directories)


class CommitViewer:
    """
    This class allows us to view the repository files and directories
    from the perspective of a specific git commit (i.e: version).

    i.e: if in the latest commit we removed a file called `a.txt`, we will
    see it while initializing CommitViewer with commit=HEAD~1, but we will
    not see `a.txt` if we initialize the CommitViewer with commit=HEAD.

    The viewer also allows us to filter out certain paths of the commit tree.
    """
    def __init__(self, commit: Commit):
        """[summary]

        Args:
            commit (Commit): the commit that defines the perspective (or lens)
                through which we look at the repo filesystem. i.e: the commit
                that defines the "checkout".
        """
        self._repo: Repo = commit.repo
        self._commit = commit
        self._root = commit.tree

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        pass

    def nodes(self, filter: Optional[NodeFilter] = None) -> Generator[VersionedNode, None, None]:
        """
        a generator yielding all the nodes (files and directories) found in
        the repository for the current commit, after applying the filter.

        Args:
            filter (Optional[NodeFilter]): an optional predicate to filter only specific nodes.

        Yields:
            the next node found (only for nodes passing the filter).
        """
        nodes_generator = self._nodes_in_tree(self._root)
        if filter is None:
            yield from nodes_generator
        else:
            for node in nodes_generator:
                if filter(node):
                    yield node

    def files(self, filter: Optional[FileFilter] = None) -> Generator[VersionedFile, None, None]:
        """
        a generator yielding all the files found in the repository
        for the current commit, after applying the filter.

        Args:
            filter (Optional[FileFilter]): an optional predicate to filter only specific files.

        Yields:
            the next file found (only for files passing the filter).
        """
        files_generator = self.nodes(lambda node: isinstance(node, VersionedFile))
        if filter is None:
            yield from files_generator
        else:
            for f in files_generator:
                if filter(f):
                    yield f

    def directories(self, filter: Optional[DirectoryFilter] = None) -> Generator[VersionedDirectory, None, None]:
        """
        a generator yielding all the directories found in the repository
        for the current commit, after applying the filter.

        Args:
            filter (Optional[DirectoryFilter]): an optional predicate to filter only specific directories.

        Yields:
            the next directory found (only for directories passing the filter).
        """
        dir_generator = self.nodes(lambda node: isinstance(node, VersionedDirectory))
        if filter is None:
            yield from dir_generator
        else:
            for directory in dir_generator:
                if filter(directory):
                    yield directory

    @property
    def paths(self) -> List[Path]:
        """
        returns all the paths in the repo for the current commit (both files and directories)
        """
        return [node.path for node in self.nodes()]

    def exists(self, path: Path) -> bool:
        """
        checks if a certain path exists in the repo in the current commit.
        """
        return path in self.paths

    def _nodes_in_tree(self, root: Tree) -> Generator[VersionedNode, None, None]:
        """
        a generator returning all the nodes (files and directories)
        under a certain git Tree (a versioned directory).
        """
        # yield current directory
        yield VersionedDirectory(root, self._commit)
        # yield files under current directory
        for blob in root.blobs:
            yield VersionedFile(blob, self._commit)
        # yield subdirectories (and their children etc) under current directory
        for tree in root.trees:
            yield from self._nodes_in_tree(tree)




from typing import List
from pathlib import Path

from opal_common.paths import PathUtils

POLICY_PREFIX = "policy:"

def policy_topics(paths: List[Path]) -> List[str]:
    """
    prefixes a list of directories with the policy topic prefix.
    """
    return ["{}{}".format(POLICY_PREFIX, str(path)) for path in paths]

def remove_prefix(topic: str, prefix: str = POLICY_PREFIX):
    """
    removes the policy topic prefix to get the path (directory)
    encoded in the topic
    """
    if topic.startswith(prefix):
        return topic[len(prefix):]
    return topic

def pubsub_topics_from_directories(dirs: List[str]) -> List[str]:
    """
    converts a list of directories on the policy repository that the
    client wants to subscribe to into a list of topics. this method
    also ensures the client only subscribes to non-intersecting directories
    by dedupping directories that are decendents of one another.
    """
    policy_directories = PathUtils.non_intersecting_directories([Path(d) for d in dirs])
    return policy_topics(policy_directories)


import asyncio
from typing import Any, List

from fastapi_websocket_pubsub import PubSubEndpoint, PubSubClient, TopicList
from opal_common.logger import logger


class TopicPublisher:
    """
    abstract publisher, base class for client side and server side publisher
    """
    def __init__(self):
        """
        inits the publisher's asyncio tasks list
        """
        self._tasks: List[asyncio.Task] = []

    def publish(self, topics: TopicList, data: Any = None):
        raise NotImplementedError()

    async def __aenter__(self):
        self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()

    def start(self):
        """
        starts the publisher
        """
        logger.info("started topic publisher")

    async def stop(self):
        """
        stops the publisher (cancels any running publishing tasks)
        """
        logger.info("stopping topic publisher")
        for task in self._tasks:
            if not task.done():
                task.cancel()
        await asyncio.gather(*self._tasks, return_exceptions=True)


class ServerSideTopicPublisher(TopicPublisher):
    """
    A simple wrapper around a PubSubEndpoint that exposes publish().
    """
    def __init__(self, endpoint: PubSubEndpoint):
        """inits the publisher.

        Args:
            endpoint (PubSubEndpoint): a pub/sub endpoint
        """
        self._endpoint = endpoint
        super().__init__()

    def publish(self, topics: TopicList, data: Any = None):
        self._tasks.append(asyncio.create_task(self._endpoint.publish(topics=topics, data=data)))


class ClientSideTopicPublisher(TopicPublisher):
    """
    A simple wrapper around a PubSubClient that exposes publish().
    Provides start() and stop() shortcuts that helps treat this client
    as a separate "process" or task that runs in the background.
    """
    def __init__(self, client: PubSubClient, server_uri: str):
        """inits the publisher.

        Args:
            client (PubSubClient): a configured not-yet-started pub sub client
            server_uri (str): the URI of the pub sub server we publish to
        """
        self._client = client
        self._server_uri = server_uri
        super().__init__()

    def start(self):
        """
        starts the pub/sub client as a background asyncio task.
        the client will attempt to connect to the pubsub server until successful.
        """
        super().start()
        self._client.start_client(f"{self._server_uri}")

    async def stop(self):
        """
        stops the pubsub client, and cancels any publishing tasks.
        """
        await self._client.disconnect()
        await super().stop()

    async def wait_until_done(self):
        """
        When the publisher is a used as a context manager, this method waits until
        the client is done (i.e: terminated) to prevent exiting the context.
        """
        return await self._client.wait_until_done()

    def publish(self, topics: TopicList, data: Any = None):
        """
        publish a message by launching a background task on the event loop.

        Args:
            topics (TopicList): a list of topics to publish the message to
            data (Any): optional data to publish as part of the message
        """
        self._tasks.append(asyncio.create_task(self._publish(topics=topics, data=data)))

    async def _publish(self, topics: TopicList, data: Any = None) -> bool:
        """
        Do not trigger directly, must be triggered via publish()
        in order to run as a monitored background asyncio task.
        """
        await self._client.wait_until_ready()
        return await self._client.publish(topics, data)


import asyncio
from typing import Coroutine

from ..events import  FetchEvent
from ..fetcher_register import FetcherRegister
from ..logger import get_logger
from .base_fetching_engine import BaseFetchingEngine

logger = get_logger("fetch_worker")

async def fetch_worker(queue: asyncio.Queue, engine):
    """
    The worker task performing items added to the Engine's Queue

    Args:
        queue (asyncio.Queue): The Queue 
        engine (BaseFetchingEngine): The engine itself
    """
    engine: BaseFetchingEngine
    register: FetcherRegister = engine.register
    while True:
        # types
        event: FetchEvent
        callback: Coroutine
        # get a event from the queue
        event, callback = await queue.get()
        # take care of it
        try:
            # get fetcher for the event
            fetcher = register.get_fetcher_for_event(event)
            # fetch
            async with fetcher:
                res = await fetcher.fetch()
                data = await fetcher.process(res)
            # callback to event owner
            try:
                await callback(data)
            except Exception as err:
                logger.exception(f"Fetcher callback - {callback} failed")
                await engine._on_failure(err, event)        
        except Exception as err:
            logger.exception("Failed to process fetch event")
            await engine._on_failure(err, event)
        finally:
            # Notify the queue that the "work item" has been processed.
            queue.task_done()



"""
Easy Python configuration interface built on rop of python-decouple.
Adding typing support and parsing with Pydantic and Enum
"""

from typing import TypeVar, Optional, Any
from functools import partial
from pydantic import BaseModel
from decouple import config, Csv, text_type, undefined, UndefinedValueError
import string

from opal_common.authentication.types import EncryptionKeyFormat, PrivateKey, PublicKey
from opal_common.authentication.casting import cast_private_key, cast_public_key

def cast_boolean(value):
    """
    Parse an entry as a boolean.
     - all variations of "true" and 1 are treated as True
     - all variations of "false" and 0 are treated as False
    """
    if isinstance(value, bool):
        return value
    elif isinstance(value, str):
        value = value.lower()
        if (value == "true" or value == "1"):
            return True
        elif value == "false" or value == "0":
            return False
        else:
            raise UndefinedValueError(f"{value} - is not a valid boolean")
    else:
        raise UndefinedValueError(f"{value} - is not a valid boolean")

def cast_pydantic(model:BaseModel):
    def cast_pydantic_by_model(value):
        if isinstance(value, str):
            return model.parse_raw(value)
        else:
            return model.parse_obj(value)
    return cast_pydantic_by_model


EnumT = TypeVar("EnumT")
T = TypeVar("T", bound=BaseModel)


class Confi:
    """
    Interface to create typed configuration entries
    """


    def __init__(self, prefix="") -> None:
        """

        Args:
            prefix (str, optional): Prefix to add to all env-var keys. Defaults to "".
        """
        self._prefix = prefix

    def _prefix_key(self, key):
        return f"{self._prefix}{key}"

    def str(self, key, default=undefined, description=None, **kwargs) -> str:
        return config(self._prefix_key(key), default=default, **kwargs)

    def int(self, key, default=undefined, description=None, **kwargs) -> int:
        return config(self._prefix_key(key), default=default, cast=int, **kwargs)

    def bool(self, key, default=undefined, description=None, **kwargs) -> bool:
        return config(self._prefix_key(key), default=default, cast=cast_boolean, **kwargs)

    def float(self, key, default=undefined, description=None, **kwargs) -> float:
        return config(self._prefix_key(key), default=default, cast=float, **kwargs)

    def list(self, key, default=undefined, sub_cast=text_type, delimiter=",", strip=string.whitespace, description=None, **kwargs) -> list:
        try:
            return config(self._prefix_key(key), cast=Csv(cast=sub_cast, delimiter=delimiter, strip=strip), **kwargs)
        except:
            if default is undefined:
                raise
            return default

    def model(self, key, model_type:T, default=undefined, description=None, **kwargs) -> T:
        """
        Parse a config using a Pydantic model
        """
        return config(self._prefix_key(key), default=default, cast=cast_pydantic(model_type), **kwargs)

    def enum(self, key, enum_type: EnumT, default=undefined, description=None, **kwargs) -> EnumT:
        try:
            return config(self._prefix_key(key), default=default, cast=enum_type, **kwargs)
        except:
            if default is undefined:
                raise
            return default

    def private_key(
        self,
        key: str,
        default: Any = undefined,
        description: str = None,
        key_format: Optional[EncryptionKeyFormat] = None,
        passphrase: Optional[str] = None,
        **kwargs
    ) -> Optional[PrivateKey]:
        """
        parse a cryptographic private key from env vars
        """
        cast_key = partial(cast_private_key, key_format=key_format, passphrase=passphrase)
        return config(self._prefix_key(key), default=default, cast=cast_key, **kwargs)

    def public_key(
        self,
        key: str,
        default: Any = undefined,
        description: str = None,
        key_format: Optional[EncryptionKeyFormat] = None,
        **kwargs
    ) -> Optional[PublicKey]:
        """
        parse a cryptographic public key from env vars
        """
        cast_key = partial(cast_public_key, key_format=key_format)
        return config(self._prefix_key(key), default=default, cast=cast_key, **kwargs)


# default parser
confi = Confi()



from typing import Tuple

from git import Repo, Head, Remote
from git.objects.commit import Commit
from tenacity import retry, wait_fixed, stop_after_attempt

from opal_common.logger import logger
from opal_common.git.exceptions import GitFailed


class BranchTracker:
    """
    tracks the state of a git branch (hash at branch HEAD).
    can also perform git pull and detect if the hash changed.
    """

    DEFAULT_RETRY_CONFIG = {
        'wait': wait_fixed(3),
        'stop': stop_after_attempt(2)
    }

    def __init__(
        self,
        repo: Repo,
        branch_name: str = "master",
        remote_name: str = "origin",
        retry_config = None,
    ):
        """[summary]

        Args:
            repo (Repo): a git repo in which we want to track the latest commit of a branch
            branch_name (str): the branch we want to track
            remote_name (str): the remote in which the branch upstream is located
            retry_config (dict): Tenacity.retry config (@see https://tenacity.readthedocs.io/en/latest/api.html#retry-main-api)
        """
        self._repo = repo
        self._branch_name = branch_name
        self._remote_name = remote_name
        self._retry_config = retry_config if retry_config is not None else self.DEFAULT_RETRY_CONFIG

        self._save_latest_commit_as_prev_commit()

    @property
    def repo(self) -> Repo:
        """
        the repo we are tracking
        """
        return self._repo

    def pull(self) -> Tuple[bool, Commit, Commit]:
        """
        git pulls from tracked remote.

        Returns:
            pull_result (bool, Commit, Commit): a tuple consisting of:
                has_changes (bool): whether the remote had new commits on our tracked branch
                prev (Commit): the previous (before the pull) top-most commit on the tracked branch
                latest (Commit): the new top-most (latest) commit on the tracked branch
        """
        self._pull()

        if (self.prev_commit.hexsha == self.latest_commit.hexsha):
            return False, self.prev_commit, self.prev_commit
        else:
            prev = self._prev_commit
            self._save_latest_commit_as_prev_commit()
            return True, prev, self.latest_commit

    def _pull(self):
        """
        runs git pull with retries.
        """
        attempt_pull = retry(**self._retry_config)(self.tracked_remote.pull)
        return attempt_pull()

    def _save_latest_commit_as_prev_commit(self):
        """
        saves the top of the branch as a last known commit (HEAD).
        in the next pull, we can then compare the new branch HEAD to the previous _prev_commit.
        """
        self._prev_commit = self.latest_commit

    @property
    def latest_commit(self) -> Commit:
        """
        the top commit (HEAD) of the tracked branch
        """
        return self.tracked_branch.commit

    @property
    def prev_commit(self) -> Commit:
        """
        the last previously known HEAD of the tracked branch
        """
        return self._prev_commit

    @property
    def tracked_branch(self) -> Head:
        """
        returns the tracked branch object (of type git.HEAD)
        or throws if such branch does not exist on the repo
        """
        try:
            return getattr(self._repo.heads, self._branch_name)
        except AttributeError as e:
            branches = [{'name': head.name, 'path': head.path} for head in self._repo.heads]
            logger.exception("did not find main branch: {error}, instead found: {branches_found}", error=e, branches_found=branches)
            raise GitFailed(e)

    @property
    def tracked_remote(self) -> Remote:
        """
        returns the tracked remote object (of type git.Remote)
        or throws if such branch does not exist on the repo
        """
        try:
            return getattr(self._repo.remotes, self._remote_name)
        except AttributeError as e:
            remotes = [remote.name for remote in self._repo.remotes]
            logger.exception("did not find main branch: {error}, instead found: {remotes_found}", error=e, remotes_found=remotes)
            raise GitFailed(e)



from opal_common.fetcher.logger import get_logger
from typing import Type, Optional, Dict
from .fetch_provider import BaseFetchProvider
from .providers.http_get_fetch_provider import HttpGetFetchProvider
from .events import FetchEvent
from ..config import FETCH_PROVIDER_MODULES

logger = get_logger("opal.fetcher_register")

class FetcherRegisterException(Exception):
    pass


class NoMatchingFetchProviderException(FetcherRegisterException):
    pass


class FetcherRegister:
    """
    A store for fetcher providers
    """

    # Builtin fetchers
    BASIC_CONFIG = {
        "HttpGetFetchProvider": HttpGetFetchProvider,
    }

    def __init__(self, config: Optional[Dict[str, BaseFetchProvider]] = None) -> None:
        if config is not None:
            self._config = config
        else:
            from ..emport import emport_objects_by_class
            # load fetchers
            fetchers = []
            for module_path in FETCH_PROVIDER_MODULES:
                try:
                    fetchers.extend(emport_objects_by_class(module_path, BaseFetchProvider, ["*"]))
                except:
                    logger.exception("Failed to load FetchingProvider module - {module_path}", module_path=module_path)
            self._config = {name: fetcher for name,fetcher in fetchers}
        logger.info("Fetcher Register loaded", extra={'config': self._config})

    def register_fetcher(self, name: str, fetcher_class: Type[BaseFetchProvider]):
        self._config[name] = fetcher_class

    def get_fetcher(self, name: str, event: FetchEvent) -> BaseFetchProvider:
        """
        Init a fetcher instance from a registered fetcher class name

        Args:
            name (str): Name of a registered fetcher
            event (FetchEvent): Event used to configure the fetcher

        Returns:
            BaseFetchProvider: A fetcher instance
        """
        provider_class = self._config.get(name, None)
        if provider_class is None:
            raise NoMatchingFetchProviderException(f"Couldn't find a match for - {name} , {event}")
        fetcher = provider_class(event)
        if event.retry is not None:
            fetcher.set_retry_config(event.retry)
        return fetcher

    def get_fetcher_for_event(self, event: FetchEvent) -> BaseFetchProvider:
        """
        Same as get_fetcher, using the event information to deduce the fetcher class

        Args:
            event (FetchEvent): Event used to choose and configure the fetcher

        Returns:
            BaseFetchProvider: A fetcher instance
        """
        return self.get_fetcher(event.fetcher, event)



"""
Simple HTTP get data fetcher using requests
supports 
"""

from aiohttp import ClientResponse, ClientSession
from ..fetch_provider import BaseFetchProvider
from ..events import FetcherConfig, FetchEvent
from ..logger import get_logger

logger = get_logger("http_get_fetch_provider")

class HttpGetFetcherConfig(FetcherConfig):
    """
    Config for HttpGetFetchProvider's Adding HTTP headers 
    """
    headers: dict = None
    is_json: bool = True
    process_data: bool = True

class HttpGetFetchEvent(FetchEvent):
    fetcher: str = "HttpGetFetchProvider"
    config: HttpGetFetcherConfig = None

class HttpGetFetchProvider(BaseFetchProvider):

    def __init__(self, event: HttpGetFetchEvent) -> None:
        self._event: HttpGetFetchEvent
        if event.config is None:
            event.config = HttpGetFetcherConfig()
        super().__init__(event)
        self._session = None

    def parse_event(self, event: FetchEvent) -> HttpGetFetchEvent:
        return HttpGetFetchEvent(**event.dict(exclude={"config"}), config=event.config)   

    async def __aenter__(self):
        headers = {}
        if self._event.config.headers is not None:
            headers = self._event.config.headers        
        self._session = await ClientSession(headers=headers).__aenter__()
        return self 

    async def __aexit__(self, exc_type=None, exc_val=None, tb=None):
        await self._session.__aexit__(exc_type=exc_type, exc_val=exc_val, exc_tb=tb)

    async def _fetch_(self):
        logger.debug(f"{self.__class__.__name__} fetching from {self._url}")
        
        result = await self._session.get(self._url)
        return result

    async def _process_(self, res: ClientResponse):
        # if we are asked to process the data before return it
        if self._event.config.process_data:
            # if data is JSON
            if self._event.config.is_json:
                data = await res.json()
            else:
                data = await res.text() 
            return data
        # return raw result
        else:
            return res



from logging import basicConfig
from typing import List, Dict, Union
from pydantic import BaseModel, Field

from opal_common.fetcher.events import FetcherConfig


class DataSourceEntry(BaseModel):
    """
    Data source configuration - where client's should retrive data from and how they should store it
    """
    # How to obtain the data
    url:str = Field(..., description="Url source to query for data")
    config:dict =  Field(None, description="Suggested fetcher configuration (e.g. auth or method) to fetch data with")
    # How to catalog data
    topics:List[str] = Field(None, description="topics the data applies to")
    # How to save the data
    # see https://www.openpolicyagent.org/docs/latest/rest-api/#data-api path is the path nested under <OPA_SERVER>/<version>/data
    dst_path:str = Field("", description="OPA data api path to store the document at")
    save_method:str = Field("PUT", description="Method used to write into OPA - PUT/PATCH")

class DataSourceConfig(BaseModel):
    """
    DataSources used as OPAL-server configuration
    """
    entries: List[DataSourceEntry] = Field(..., description="list of data sources and how to fetch from them")

class DataUpdate(BaseModel):
    """
    DataSources used as OPAL-server configuration
    Data update sent to clients
    """
    entries: List[DataSourceEntry] = Field(..., description="list of related updates the OPAL client should perform")
    reason: str = Field(None, description="Reason for triggering the update")






import re

from typing import Optional

# This regex matches the package declaration at the top of a valid .rego file
REGO_PACKAGE_DECLARATION = re.compile(r'^package\s+([a-zA-Z0-9\.\"\[\]]+)$')


def get_rego_package(contents: str) -> Optional[str]:
    """
    try to parse the package name from rego file contents

    return None if failed to parse (probably invalid .rego file)
    """
    lines = contents.splitlines()
    for line in lines:
        match = REGO_PACKAGE_DECLARATION.match(line)
        if match is not None:
            return match.group(1)
    return None



from datetime import datetime, timedelta
from typing import Optional
from uuid import UUID

import jwt
from fastapi import HTTPException, status
from jwt.algorithms import Algorithm, get_default_algorithms
from jwt.api_jwk import PyJWK

from opal_common.authentication.types import JWTAlgorithm, JWTClaims, PrivateKey, PublicKey
from opal_common.logger import logger


class Unauthorized(HTTPException):
    """
    HTTP 401 Unauthorized exception.
    """
    def __init__(self, description="Bearer token is not valid!", **kwargs):
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail={
                "error": description,
                **kwargs
            },
            headers={"WWW-Authenticate": "Bearer"}
        )


class InvalidJWTCryptoKeysException(Exception):
    """
    raised when JWT signer provided with invalid crypto keys
    """
    pass

class JWTSigner:
    """
    given cryptographic keys, signs and verifies jwt tokens.
    """
    def __init__(
        self,
        private_key: Optional[PrivateKey],
        public_key: Optional[PublicKey],
        algorithm: JWTAlgorithm,
        audience: str,
        issuer: str,
    ):
        """
        inits the signer if and only if the keys provided to __init__
        were generate together are are valid. otherwise will throw.

        JWT signer can be initialized with empty keys (None),
        in which case signer.enabled == False.

        This allows opal to run both in secure mode (which keys, requires jwt authentication)
        and in insecure mode (good for development and running locally).

        Args:
            private_key (PrivateKey): a valid private key or None
            public_key (PublicKey): a valid public key or None
            algorithm (JWTAlgorithm): the jwt algorithm to use
                (possible values: https://pyjwt.readthedocs.io/en/stable/algorithms.html)
            audience (string): the value for the aud claim: https://tools.ietf.org/html/rfc7519#section-4.1.3
            issuer (string): the value for the iss claim: https://tools.ietf.org/html/rfc7519#section-4.1.1
        """
        self._private_key = private_key
        self._public_key = public_key
        self._algorithm: str = algorithm.value
        self._audience = audience
        self._issuer = issuer
        self._enabled = True
        self._verify_crypto_keys()

    def _verify_crypto_keys(self):
        """
        verifies whether or not valid crypto keys were provided to the signer.
        if both keys are valid, encodes and decodes a JWT to make sure the keys match.

        if both private and public keys are valid and are matching => signer is enabled
        if both private and public keys are None => signer is disabled (self.enabled == False)
        if only one key is valid/not-None => throws ValueError
        any other case => throws ValueError
        """
        if self._private_key is not None and self._public_key is not None:
            # both keys provided, let's make sure these keys were generated correctly
            token = jwt.encode({"some": "payload"}, self._private_key, algorithm=self._algorithm)
            try:
                jwt.decode(token, self._public_key, algorithms=[self._algorithm])
            except jwt.PyJWTError as exc:
                logger.info("JWT Signer key verification failed with error: {err}", err=exc)
                raise InvalidJWTCryptoKeysException("private key and public key do not match!") from exc
            # save jwk
            self._jwk: PyJWK = PyJWK.from_json(self.get_jwk(), algorithm=self._algorithm)
        elif (self._private_key != self._public_key) and (self._private_key is None or self._public_key is None):
            raise ValueError("JWT Signer not valid, only one of private key / public key pair was provided!")
        elif self._private_key is None and self._public_key is None:
            # valid situation, running in dev mode and api security is off
            self._enabled = False
            logger.info("OPAL was not provided with JWT encryption keys, cannot verify api requests!")
        else:
            raise ValueError("Invalid JWT Signer input!")

    def get_jwk(self) -> str:
        """
        returns the jwk json contents
        """
        algorithm: Optional[Algorithm] = get_default_algorithms().get(self._algorithm)
        if algorithm is None:
            raise ValueError(f"invalid jwt algorithm: {self._algorithm}")
        return algorithm.to_jwk(self._public_key)

    @property
    def enabled(self):
        """
        whether or not the signer has valid cryptographic keys
        """
        return self._enabled

    def verify(self, token: str) -> JWTClaims:
        """
        verifies a JWT token is valid.
        if valid returns dict with jwt claims, otherwise throws.
        """
        try:
            return jwt.decode(
                token,
                self._public_key,
                algorithms=[self._algorithm],
                audience=self._audience,
                issuer=self._issuer
            )
        except jwt.ExpiredSignatureError:
            raise Unauthorized(token=token, description="Access token is expired")
        except jwt.InvalidAudienceError:
            raise Unauthorized(token=token, description="Invalid access token: invalid audience claim")
        except jwt.InvalidIssuerError:
            raise Unauthorized(token=token, description="Invalid access token: invalid issuer claim")
        except jwt.DecodeError:
            raise Unauthorized(token=token, description="Could not decode access token")
        except Exception:
            raise Unauthorized(token=token, description="Unknown JWT error")

    def sign(self, sub: UUID, token_lifetime: timedelta, custom_claims: dict = {}) -> str:
        payload = {}
        issued_at = datetime.utcnow()
        expire_at = issued_at + token_lifetime
        payload = {
            "iat": issued_at,
            "exp": expire_at,
            "aud": self._audience,
            "iss": self._issuer,
            "sub": sub.hex,
        }
        if custom_claims:
            payload.update(custom_claims)

        headers = {}
        if self._jwk.key_id is not None:
            headers = {"kid": self._jwk.key_id}
        return jwt.encode(payload, self._private_key, algorithm=self._algorithm, headers=headers)
